import * as i0 from '@angular/core';
import { Injectable, EventEmitter, Component, Input, Output, NgModule } from '@angular/core';
import { Subject } from 'rxjs';
import * as d3 from 'd3';
import { forceSimulation, forceManyBody, forceCollide, forceLink } from 'd3-force';
import { scaleLog } from 'd3-scale';
import * as uuid from 'uuid';
import * as d3SvgToPng from 'd3-svg-to-png';
import * as i2 from '@angular/common';
import { CommonModule } from '@angular/common';

class NgxFishboneDiagramService {
    constructor() {
        this.restartSource = new Subject();
        this.restartRequest$ = this.restartSource.asObservable();
    }
    restart() {
        this.restartSource.next(true);
    }
}
NgxFishboneDiagramService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: NgxFishboneDiagramService, deps: [], target: i0.ɵɵFactoryTarget.Injectable });
NgxFishboneDiagramService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: NgxFishboneDiagramService, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: NgxFishboneDiagramService, decorators: [{
            type: Injectable,
            args: [{
                    providedIn: 'root'
                }]
        }], ctorParameters: function () { return []; } });

class NgxFishboneDiagramComponent {
    constructor(svc) {
        this.svc = svc;
        this.showDowloadButton = false;
        this.btnClass = 'downloadButton';
        this.btnText = 'Download PNG';
        this.selected = new EventEmitter();
        this.nodeIdx = 0;
        this.margin = 100;
        this.nodes = new Array();
        this.links = new Array();
        this.width = 1280;
        this.height = 600;
        this.svc.restartRequest$.subscribe((req) => {
            if (req) {
                this.restart();
            }
        });
    }
    ngOnInit() {
        this.force = forceSimulation(this.nodes)
            .force("charge", forceManyBody().strength(-10))
            .force("collision", forceCollide(15))
            .force('link', forceLink(this.links).distance(this.linkDistance))
            .on("end", () => this.simulationDone())
            .on("tick", () => this.tick());
    }
    ngOnChanges(changes) {
        // check if "data" input has changed
        if (changes.data) {
            if (!this.svg) {
                /* TODO: A more angular way to do this is to use a ViewChild. */
                this.svg = d3.select("#d3Container")
                    .append('svg')
                    .attr('width', this.width)
                    .attr('height', this.height)
                    .call(this.defaultArrow);
            }
            this.nodes = [];
            this.links = [];
            if (changes.data.currentValue === null) {
                this.clear();
            }
            else {
                this.buildNodes(changes.data.currentValue);
                this.restart();
            }
        }
    }
    clear() {
        this.svg.remove();
        this.node.remove();
        this.link.remove();
        this.root.remove();
        this.nodes = [];
        this.links = [];
        this.force.nodes(this.nodes);
    }
    restart() {
        this.setupNodes();
        if (this.force) {
            this.force.stop(); // this is just in case. ideally, the force system either doesn't exist here or is already stopped.
            this.force.nodes(this.nodes, (d) => { return d.uuid; });
            this.force.force('link', forceLink(this.links).distance(this.linkDistance));
            this.force.alpha(1).restart();
        }
    }
    setupNodes() {
        /* setup the nodes */
        this.node = this.svg.selectAll(".node")
            .data(this.nodes, (d) => { return d.uuid; });
        this.node.enter()
            .append("g")
            .attr("class", (d) => { return d.root ? "node root" : "node"; })
            .on("click", (d, i) => { this.nodeClicked(d, i); })
            .append("text");
        /* find all text nodes and add the actual text to it. */
        this.svg.selectAll("text")
            .attr("class", (d) => "label-" + d.depth)
            .attr("text-anchor", (d) => { return !d.depth ? "start" : d.horizontal ? "end" : "middle"; })
            .attr("dy", (d) => { return d.horizontal ? ".35em" : d.region === 1 ? "1em" : "-0.2em"; })
            .text((d) => { return d.name; });
        this.svg.selectAll("text")
            .call(this.wrap, 100);
        this.node.exit().remove();
        /* setup the links */
        this.link = this.svg.selectAll('.link')
            .data(this.links);
        this.link.enter()
            .append('line')
            .attr("class", (d) => { return "link link-" + d.depth; })
            .attr("marker-end", (d) => { return d.arrow ? "url(#arrow)" : null; });
        this.link.exit()
            .remove();
        this.root = d3.select(".root").node();
    }
    /**
     *  buildNodes is a function to rebuild the node list and pad all the nodes with layout directives.
     *  For the layout and forces to work properly, we use a variety of node properties like horizontal,
     *  vertical, region, depth, tail among others.
     * */
    buildNodes(node) {
        /* don't add a node that already exists in the list of nodes. */
        const idx = this.nodes.findIndex((val) => { return val.uuid === node.uuid; });
        if (idx === -1) {
            this.nodes.push(node);
        }
        var cx = 0;
        let between = [node, node.connector];
        let nodeLinks = [{
                source: node,
                target: node.connector,
                arrow: true,
                depth: node.depth || 0
            }];
        let prev;
        let childLinkCount;
        if (!node.parent) {
            /* don't add a tail, if one already exists in the list. */
            const tailIdx = this.nodes.findIndex((val) => { return val.tail; });
            if (tailIdx === -1) {
                this.nodes.push(prev = { tail: true, uuid: uuid.v4() });
            }
            else {
                prev = this.nodes[tailIdx];
            }
            between = [prev, node];
            nodeLinks[0].source = prev;
            nodeLinks[0].target = node;
            node.horizontal = true;
            node.vertical = false;
            node.depth = 0;
            node.root = true;
            node.totalLinks = [];
        }
        else {
            node.connector.maxChildIdx = 0;
            node.connector.totalLinks = [];
        }
        node.linkCount = 1;
        (node.children || []).forEach((child, idx) => {
            child.parent = node;
            child.depth = (node.depth || 0) + 1;
            child.childIdx = idx;
            child.region = node.region ? node.region : (idx & 1 ? 1 : -1);
            child.horizontal = !node.horizontal;
            child.vertical = !node.vertical;
            if (node.root && prev && !prev.tail) {
                this.nodes.push(child.connector = {
                    between: between,
                    childIdx: prev.childIdx,
                    uuid: uuid.v4()
                });
                prev = null;
            }
            else {
                this.nodes.push(prev = child.connector = { between: between, childIdx: cx++, uuid: uuid.v4() });
            }
            nodeLinks.push({
                source: child,
                target: child.connector,
                depth: child.depth,
                arrow: true
            });
            /* recurse capturing number of links created */
            childLinkCount = this.buildNodes(child);
            node.linkCount += childLinkCount;
            between[1].totalLinks.push(childLinkCount);
        });
        between[1].maxChildIdx = cx;
        nodeLinks.forEach((link) => {
            let lidx = this.links.findIndex((val) => {
                return val.source.uuid === link.source.uuid && val.target.uuid === link.target.uuid;
            });
            if (lidx === -1) {
                this.links.push(link);
            }
        });
        //this.links.unshift(...nodeLinks);
        /* the number of links created byt this node and its children...
           TODO: use `linkCount` and/instead of `childIdx` for spacing */
        return node.linkCount;
    }
    /**
     * tick is the actual force layout method. Force is a simulation that happens in
     * steps. At each step, a `tick` event is generated. We connect the event to this
     * method. In this method, we compute the locations of each node based on the forces
     * and position them accordingly.
     * */
    tick() {
        let k = this.force.alpha() * 0.8;
        this.nodes.forEach((n) => this.calculateXY(n, k));
        d3.selectAll('.node').attr("transform", function (d) {
            return "translate(" + d.x + "," + d.y + ")";
        });
        d3.selectAll('.link')
            .attr("x1", (d) => { if (d.source)
            return d.source.x; })
            .attr("x2", (d) => { if (d.target)
            return d.target.x; })
            .attr("y1", (d) => { if (d.source)
            return d.source.y; })
            .attr("y2", (d) => { if (d.target)
            return d.target.y; });
    }
    calculateXY(d, k) {
        this.root = d3.select('.root').node();
        if (!this.root) {
            return;
        }
        /* handle the middle... could probably store the root width... */
        if (d.root) {
            d.x = this.width - (this.margin + this.root.getBBox().width);
        }
        if (d.tail) {
            d.x = this.margin;
            d.y = this.height / 2;
        }
        /* put the first-generation items at the top and bottom */
        if (d.depth === 1) {
            d.y = d.region === -1 ? this.margin : (this.height - this.margin);
            d.x -= 10 * k;
        }
        /* vertically-oriented tend towards the top and bottom of the page */
        if (d.vertical) {
            d.y += k * d.region;
        }
        /* everything tends to the left */
        if (d.depth) {
            d.x -= k;
        }
        /* position synthetic nodes at evently-spaced intervals...
           TODO: do something based on the calculated size of each branch
           since we don't have individual links anymore */
        let a, b;
        if (d.between) {
            a = d.between[0];
            b = d.between[1];
            d.x = b.x - (1 + d.childIdx) * (b.x - a.x) / (b.maxChildIdx + 1);
            d.y = b.y - (1 + d.childIdx) * (b.y - a.y) / (b.maxChildIdx + 1);
        }
    }
    defaultArrow(svg) {
        /* creates an svg:defs and marker with an arrow if needed...
           really just an example, as they aren't very flexible */
        var defs = svg.selectAll("defs").data([1]);
        defs.enter()
            .append("defs");
        /* create the arrows */
        svg.selectAll("marker#arrow")
            .data([1])
            .enter()
            .append("marker")
            .attr("id", "arrow")
            .attr("viewBox", "0 -5 10 10")
            .attr("refX", 10)
            .attr("refY", 0)
            .attr("markerWidth", 10)
            .attr("markerHeight", 10)
            .attr("orient", "auto")
            .append("path")
            .attr("d", "M0,-5L10,0L0,5");
    }
    linkDistance(l) {
        const linkScale = scaleLog().domain([1, 5]).range([70, 10]);
        return (l.target.maxChildIdx + 1) * linkScale(l.depth + 1);
    }
    nodeClicked(d, i) {
        if (this.force) {
            this.force.stop();
        }
        this.selected.emit(i.uuid);
    }
    downloadImage() {
        d3SvgToPng.default('svg', 'fishboneDiagram', { download: true, format: 'png' });
    }
    simulationDone() {
        console.info("layout complete.");
        console.log(this.nodes.length + ' nodes and ' + this.links.length + ' links');
    }
    /*
     * text wrap function for use with SVG.
     * https://bl.ocks.org/mbostock/7555321
     */
    wrap(selectionList, width) {
        selectionList.each(function (selection, idx, selList) {
            let text = selection.name;
            let domNode = d3.select(selList[idx]);
            let dy = parseFloat(domNode.attr('dy'));
            let y = domNode.attr('y');
            let words = [];
            if (text) {
                words = text.split(/\s+/).reverse();
            }
            let word;
            let line = [];
            let lineNumber = 0;
            // let lineHeight = calculatePosition(selection.depth, selection.region, lineNumber);
            let tspan = domNode.text(null).append("tspan").attr("x", 0).attr("y", y).attr("dy", `${dy}em`);
            while (word = words.pop()) {
                line.push(word);
                tspan.text(line.join(" "));
                let tspanNode = tspan.node();
                let t = line.join(" ");
                if (tspanNode && tspanNode.getComputedTextLength() > width) {
                    line.pop();
                    tspan.text(line.join(" "));
                    line = [word];
                    let _dy = calculatePosition(selection.depth, selection.region, ++lineNumber);
                    tspan = domNode.append("tspan").attr("x", 0).attr("y", y).attr("dy", `${_dy}em`).text(word);
                }
            }
        });
    }
}
NgxFishboneDiagramComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: NgxFishboneDiagramComponent, deps: [{ token: NgxFishboneDiagramService }], target: i0.ɵɵFactoryTarget.Component });
NgxFishboneDiagramComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "12.0.0", version: "13.3.11", type: NgxFishboneDiagramComponent, selector: "ngx-fishbone-diagram", inputs: { data: "data", showDowloadButton: "showDowloadButton", btnClass: "btnClass", btnText: "btnText" }, outputs: { selected: "selected" }, usesOnChanges: true, ngImport: i0, template: "\n<button *ngIf=\"showDowloadButton\" (click)=\"downloadImage()\" [ngClass]=\"btnClass\">{{ btnText }}</button>\n<div #d3Container id=\"d3Container\" class=\"d3-container\" [style.width.px]=\"width\" [style.height.px]=\"height\"></div>\n", styles: [".d3-container{margin:0 auto}.downloadButton{float:right;margin-right:10%;background-color:#fff;border:gray 1px solid;border-radius:4px;padding:4px;cursor:pointer}::ng-deep .node{cursor:pointer}::ng-deep text.label-0{font-size:1.1em}::ng-deep text.label-1{font-size:1em;fill:#111}::ng-deep text.label-2{font-size:.9em;fill:#444}::ng-deep text.label-3{font-size:.8em;fill:#888}::ng-deep text.label-4{font-size:.7em;fill:#aaa}::ng-deep .link-0{stroke:#000;stroke-width:2px}::ng-deep .link-1{stroke:#333;stroke-width:1px}::ng-deep .link-2{stroke:#666;stroke-width:1px}::ng-deep .link-3{stroke:#666;stroke-width:1px}::ng-deep .link-4{stroke:#666;stroke-width:1px}\n"], directives: [{ type: i2.NgIf, selector: "[ngIf]", inputs: ["ngIf", "ngIfThen", "ngIfElse"] }, { type: i2.NgClass, selector: "[ngClass]", inputs: ["class", "ngClass"] }] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: NgxFishboneDiagramComponent, decorators: [{
            type: Component,
            args: [{ selector: 'ngx-fishbone-diagram', template: "\n<button *ngIf=\"showDowloadButton\" (click)=\"downloadImage()\" [ngClass]=\"btnClass\">{{ btnText }}</button>\n<div #d3Container id=\"d3Container\" class=\"d3-container\" [style.width.px]=\"width\" [style.height.px]=\"height\"></div>\n", styles: [".d3-container{margin:0 auto}.downloadButton{float:right;margin-right:10%;background-color:#fff;border:gray 1px solid;border-radius:4px;padding:4px;cursor:pointer}::ng-deep .node{cursor:pointer}::ng-deep text.label-0{font-size:1.1em}::ng-deep text.label-1{font-size:1em;fill:#111}::ng-deep text.label-2{font-size:.9em;fill:#444}::ng-deep text.label-3{font-size:.8em;fill:#888}::ng-deep text.label-4{font-size:.7em;fill:#aaa}::ng-deep .link-0{stroke:#000;stroke-width:2px}::ng-deep .link-1{stroke:#333;stroke-width:1px}::ng-deep .link-2{stroke:#666;stroke-width:1px}::ng-deep .link-3{stroke:#666;stroke-width:1px}::ng-deep .link-4{stroke:#666;stroke-width:1px}\n"] }]
        }], ctorParameters: function () { return [{ type: NgxFishboneDiagramService }]; }, propDecorators: { data: [{
                type: Input
            }], showDowloadButton: [{
                type: Input
            }], btnClass: [{
                type: Input
            }], btnText: [{
                type: Input
            }], selected: [{
                type: Output
            }] } });
function calculatePosition(depth, region, lineNumber) {
    // at depth 1, for the top half, increment by one lineNumber;
    // let lh = 1;
    // console.log(`calculateLineHeight(depth: ${depth}, region: ${region}, lineNumber: ${lineNumber}) = ${lh}`);
    return 1;
}

class NgxFishboneDiagramModule {
}
NgxFishboneDiagramModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: NgxFishboneDiagramModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgxFishboneDiagramModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: NgxFishboneDiagramModule, declarations: [NgxFishboneDiagramComponent], imports: [CommonModule], exports: [NgxFishboneDiagramComponent] });
NgxFishboneDiagramModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: NgxFishboneDiagramModule, providers: [
        NgxFishboneDiagramService
    ], imports: [[
            CommonModule
        ]] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: NgxFishboneDiagramModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [
                        NgxFishboneDiagramComponent
                    ],
                    imports: [
                        CommonModule
                    ],
                    exports: [
                        NgxFishboneDiagramComponent
                    ],
                    providers: [
                        NgxFishboneDiagramService
                    ]
                }]
        }] });

/*
 * Public API Surface of ngx-fishbone-diagram
 */

/**
 * Generated bundle index. Do not edit.
 */

export { NgxFishboneDiagramComponent, NgxFishboneDiagramModule, NgxFishboneDiagramService };
//# sourceMappingURL=ngx-fishbone-diagram.mjs.map
